/*
 * These are the stub functions for ksim that are compiled against
 * kernel headers.
 */
#define __KERNEL__

#include <linux/autoconf.h>

#ifdef CONFIG_SMP
#define __SMP__
#endif

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/timer.h>
#include <linux/genhd.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/mman.h>
#include <linux/hdreg.h>
#include <linux/signal.h>
#include <asm/semaphore-helper.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>

#if 0
#include <stdio.h>
#include <sys/times.h>
#endif

#include <linux/sched.h>
#include <linux/kernel_stat.h>
#include <linux/mm.h>
#include <linux/swapctl.h>
#include <asm/smp.h>

#define SCSI_IOCTL_DOORLOCK 0x5380		/* lock the eject mechanism */
#define SCSI_IOCTL_DOORUNLOCK 0x5381		/* unlock the mechanism	  */

#define get_active_bhs()	(bh_mask & bh_active)
#define clear_active_bhs(x)	atomic_clear_mask((x),&bh_active)

#ifdef CONFIG_SMP
volatile unsigned char active_kernel_processor = NO_PROC_ID;
unsigned char global_irq_holder = NO_PROC_ID;
#endif

spinlock_t semaphore_wake_lock = SPIN_LOCK_UNLOCKED;

spinlock_t tqueue_lock;
rwlock_t waitqueue_lock = RW_LOCK_UNLOCKED;
unsigned int local_irq_count[NR_CPUS];
unsigned int local_bh_count[NR_CPUS];
void setup_tasklist(char * crnt);

#ifdef CONFIG_SMP
struct cpuinfo_x86 cpu_data[NR_CPUS];
// #define current_cpu_data cpu_data[smp_processor_id()]
#else
struct cpuinfo_x86 boot_cpu_data = { 0, 0, 0, 0, -1, 1, 0, 0, -1 };
#endif

int idle_thread;

void __lock_kernel(void)
{
  return;
}

volatile unsigned long kernel_flag=0;             /* Kernel spinlock     */
volatile unsigned long smp_invalidate_needed = 1;  
struct proc_dir_entry  *proc_scsi = NULL;
unsigned long bh_active = 0;
unsigned long bh_mask = 0;
int bh_mask_count[32];
void (*bh_base[32])(void);
atomic_t nr_async_pages = ATOMIC_INIT(0);
int nr_free_pages = 0;
unsigned long max_mapnr = 0;
void * high_memory = NULL;
unsigned long num_physpages = 0;
static buffer_pages = 0;

#define NR_PAGES	1024	/* 4Mb "machine" */
char buffer0[(NR_PAGES+1)*4096];
char * buffer2;
char buffer[1024*4096];
mem_map_t __mem_map[NR_PAGES];
mem_map_t * mem_map = &__mem_map[0];
char * freelist[NR_PAGES];
next_free = 0;

const char lk_lockmsg[] = "lock from interrupt context at %p\n";

unsigned long __get_free_pages(int gfp_mask, unsigned long order)
{
  int i;

  if( order > 1 )
    {
      return malloc(PAGE_SIZE << order);
    }

  /*
   * Find the actual size that we need.
   */
  if( next_free > 0 )
    {
      return freelist[next_free--];
    }

  for(i=0; i < NR_PAGES; i++)
    {
      if( mem_map[i].buffers )
	{
	  if(try_to_free_buffers(&mem_map[i]) )
	    {
	      break;
	    }
	}
    }
	  
  if( next_free > 0 )
    {
      return freelist[next_free--];
    }
  return 0;
}

int __free_page(struct page *page)
{
  int index;
  /*
   * Calculate the index.
   */
  index = page - mem_map;
  freelist[++next_free] = buffer2 + (index << 12);
}

int free_pages(unsigned long addr, unsigned long order)
{
  return;
}

void add_blkdev_randomness(int major)
{
  return;
}

int ide_init(void)
{
  return 0;
}

int floppy_init(void)
{
  return 0;
}

int proc_dostring(ctl_table * a, int b, struct file * c, void * d, size_t * e)
{
        return -ENOSYS;
}

int invalidate_inodes(struct super_block * sb)
{
  return 0;
}

struct super_block * get_super(kdev_t dev)
{
  return NULL;
}

int proc_register(struct proc_dir_entry * dir, struct proc_dir_entry * dp)
{
  return 0;
}

void sync_inodes(kdev_t dev)
{
}

struct inode_operations proc_scsi_inode_operations = { NULL, };
int (* dispatch_scsi_info_ptr)(int ino, char *buffer, char **start,
				      off_t offset, int length, int inout);
int ide_xlate_1024 (kdev_t i_rdev, int xparm, const char *msg) { return 0; }
int chr_dev_init(void){ }
void console_map_init(void) { }

unsigned long swapper_pg_dir[1024];
unsigned long t1_pg_dir[1024];
unsigned long t2_pg_dir[1024];
char pg0[4096];
char empty_bad_page[4096];
char empty_bad_page_table[4096];
char empty_zero_page[4096];

/*
 * This is used to shut down the error handler thread.  It is sufficient to
 * bump the counter.
 */
int send_sig(int sig,struct task_struct * p,int priv)
{
  p->sigpending++;
  return 0;
}

void __global_cli() { return; }
void __global_sti() { return; }
unsigned long __global_save_flags() { return 0; }
void __global_restore_flags(unsigned long foob) { return; }

void __wake_up(wait_queue_head_t * p, unsigned int mode) __attribute__((regparm(3)));

void __wake_up(wait_queue_head_t * q, unsigned int mode) 
{ 
	struct list_head *tmp, *head;
	struct task_struct *p;
        if (!q)
		goto out;

	head = &q->task_list;
	tmp = head->next;
	while (tmp != head) {
		unsigned int state;
                wait_queue_t *curr = list_entry(tmp, wait_queue_t, task_list);

		tmp = tmp->next;

		p = curr->task;
		state = p->state;
		if (state & mode) {
			p->state = TASK_RUNNING;
			if (state & TASK_EXCLUSIVE)
				break;
		}
	}
out:
    return; 
}
#define MAX_TIMERS	10
struct timer_list * active_timers[MAX_TIMERS];
void add_timer(struct timer_list * timer) 
{ 
    int i;
    for(i=0; i < MAX_TIMERS; i++)
    {
	if( active_timers[i] == NULL )
	{
	    active_timers[i] = timer;
	    break;
	}
    }
    return; 
}
int  del_timer(struct timer_list * timer) 
{
    int i;
    for(i=0; i < MAX_TIMERS; i++)
    {
	if( active_timers[i] == timer )
	{
	    active_timers[i] = NULL;
	    break;
	}
    }
    return 1; 
}
void fire_next_timer()
{
    int i;
    int next_timer_index = -1;
    unsigned int next_timer = 0xffffffff;

    for(i=0; i < MAX_TIMERS; i++)
    {
	if( active_timers[i] == NULL )
	{
	    continue;
	}
	if( active_timers[i]->expires < next_timer )
	{
	    next_timer_index = i;
	    next_timer = active_timers[i]->expires;
	}
    }
    if( next_timer_index != -1 )
    {
	i = next_timer_index;
	(*active_timers[i]->function)(active_timers[i]->data);
    }
}

int request_dma(){  return 0;}
int check_region(){  return 0;}
int snarf_region(){  return 0;}
void release_resource(int class, unsigned long from, unsigned long num){ }

void disable_irq(unsigned int unused){  return;}
void enable_irq(unsigned int unused){  return;}
void free_irq(unsigned int irq, void * foob){  return;}
int request_irq(unsigned int irq, 
		void (*handler)(int, void *, struct pt_regs *),
		unsigned long flags,
		const char * device,
		void * dev_id)
{
  return 0;
}

void file_table_init(void)
{
}

void locks_remove_flock(struct file *filp)
{
}
void write_inode_now(struct inode *inode)
{
}
void insert_file_free(struct file *file)
{
}

int __fput(struct file *filp)
{
}

void __wait_on_super(struct super_block * sb)
{
}

void swap_after_unlock_page (unsigned long entry)
{
}

unsigned long timer_active;
struct timer_struct timer_table[32];

volatile unsigned long jiffies = 10000;
extern int nr_hash;
extern struct buffer_head ** hash_table;

int need_resched = 0;
int ramdisk_size = 0;
static int current_minor = 0;


struct kernel_stat kstat = {0};

struct task_struct *task[NR_TASKS];
struct task_struct * current;
int task_index = 0;
int max_reg_tasks = 0;
union  task_union init_task_union;

struct list_head super_blocks;

void do_exit(long error_code){
  exit(0);
}

void shm_no_page(unsigned long * ptr){ }
void shm_swap(int i) { }

/*static*/ void die_if_kernel(char * str, struct pt_regs * regs, long err)
{
  exit(0);
}


int in_group_p(gid_t grp) { return 0; }

struct timeval xtime;

void rd_load() { }

int floppy_change(struct buffer_head * bh){ return; }
void put_super(dev_t dev){ return; }
void sync_supers(dev_t dev){ return; }
int hd_init(int a, int b){ return a; }
int rd_init(int a, int b){ return a; }
int wp_works_ok = 1;
unsigned long intr_count = 0;
unsigned long loops_per_sec = 0;

buffer_mem_t buffer_mem = {
	3,	/* minimum percent buffer */
	10,	/* borrow percent buffer */
	30	/* maximum percent buffer */
};

freepages_t freepages = {
  48,     /* freepages.min */
  96,     /* freepages.low */
  144     /* freepages.high */
};

void free_dma(unsigned int dmar) { return; }

int irqaction(unsigned int irq,struct sigaction * sa)
{
  return -1;
}
void exit_mm(struct task_struct *tsk)
{
}

void exit_fs(struct task_struct *tsk)
{
}

void init_fifo(struct inode * inode)
{
}

static inline void run_bottom_halves(void)
{
	unsigned long active;
	void (**bh)(void);

	active = get_active_bhs();
	clear_active_bhs(active);
	bh = bh_base;
	do {
		if (active & 1)
			(*bh)();
		bh++;
		active >>= 1;
	} while (active);
}

void idle()
{
  while(1==1)
    {
      if (bh_active & bh_mask)
	{
	  run_bottom_halves();
	}
      else
	{
	    fire_next_timer();
	}
      schedule();
    }
}

void schedule()
{
    int i,j;

    while (bh_active & bh_mask)
      {
	run_bottom_halves();
      }

    for(i=0; i < max_reg_tasks; i++)
    {
        if( task[i] == NULL )
	{
	    continue;
	}

	if( (   (task[i]->state == TASK_RUNNING )
	     || (task[i]->state == TASK_INTERRUPTIBLE && task[i]->sigpending )) 
	    && (i != idle_thread) )
	{
	    int j;
	    
	    if( task[task_index] != NULL )
	    {
		j = setjmp(&task[task_index]->tss);
	    }
	    else
	    {
		j = 0;
	    }
	    if( j == 0 )
	    {
		task_index = i;
//		write(2, "Task switch\n", 11);
		longjmp(&task[i]->tss, 0);
	    }
	    else
	    {
		return;
	    }
	}
    }
    /*
     * Rats.  Transfer control to the idle thread.
     */
    {
	 int j;
	 
	 j = setjmp(&task[task_index]->tss);
	 if( j == 0 )
	 {
	      task_index = idle_thread;
	      longjmp(&task[idle_thread]->tss, 0);
	 }
	 else
	 {
	      return;
	 }
    }
    schedule_oops();
}

extern void __add_wait_queue(wait_queue_head_t * head, wait_queue_t * new)
{
	list_add(&new->task_list, &head->task_list);
}

extern void __remove_wait_queue(wait_queue_head_t * head, wait_queue_t * old)
{
	list_del(&old->task_list);
}

void interruptible_sleep_on(wait_queue_head_t * foo) __attribute__((regparm(3)));

void interruptible_sleep_on(wait_queue_head_t * q)
{
  unsigned long flags;
  wait_queue_t wait;
  init_waitqueue_entry(&wait, current);

    wait.task = current;
    task[task_index]->state = TASK_INTERRUPTIBLE;

    __add_wait_queue(q, &wait);
    schedule();
    __remove_wait_queue(q, &wait);
}

void sleep_on(wait_queue_head_t * foo) __attribute__((regparm(3)));

long sleep_on_timeout(wait_queue_head_t * q, signed long timeout)
{
  sleep_on(q);
  return 0;
}

void sleep_on(wait_queue_head_t * q)
{
  unsigned long flags;
  wait_queue_t wait;
  init_waitqueue_entry(&wait, current);

    wait.task = current;
    task[task_index]->state = TASK_INTERRUPTIBLE;

    __add_wait_queue(q, &wait);
    schedule();
    __remove_wait_queue(q, &wait);
}

void __down(struct semaphore * sem)
{ 
  struct task_struct *tsk = current;
  wait_queue_t wait;
  init_waitqueue_entry(&wait, tsk);

  tsk->state = TASK_UNINTERRUPTIBLE;

  __add_wait_queue(&sem->wait, &wait);
    /*
     * "Run" the bottom half handlers, if there is anything to do.
     */
    for (;;) 
    {
	if( sem->count.counter >= 0 )
	{
	    break;
	}
	schedule();
    }
    tsk->state = TASK_RUNNING;
    __remove_wait_queue(&sem->wait, &wait);
}

struct blk_dev_struct {
	void (*request_fn)(void);
	struct request * current_request;
};


struct blk_dev_struct blk_dev[10];

extern struct wait_queue * wait_for_request;


#if 0 /* from genhd.c */
struct gendisk *gendisk_head = NULL;
void resetup_one_dev(struct gendisk *dev, int drive){ }
#endif

void
flush_signals(struct task_struct *t)
{
}
void __up(struct semaphore *sem)
{
	wake_one_more(sem);
	wake_up(&sem->wait);
}
int __down_interruptible(struct semaphore * sem)
{
    int ret = 0;
    struct task_struct *tsk = current;
    wait_queue_t wait;
    init_waitqueue_entry(&wait, tsk);

    tsk->state = TASK_INTERRUPTIBLE;

    __add_wait_queue(&sem->wait, &wait);
    /*
     * "Run" the bottom half handlers, if there is anything to do.
     */
    for (;;) 
    {
	if( sem->count.counter >= 0 )
	{
	    break;
	}

	if (tsk->sigpending) 
	  {
	    ret = -EINTR;			/* interrupted */
	    atomic_inc(&sem->count);	/* give up on down operation */
	    break;
	  }
	schedule();
    }
    tsk->state = TASK_RUNNING;
    __remove_wait_queue(&sem->wait, &wait);
}

int __verify_write(unsigned long i, unsigned long j)
{
  return 0;
}

#define MAJOR_NR 8
#define CHR_MAJOR_NR 9
#define IO_BLOCKSIZE 1024

char test_buffer2[IO_BLOCKSIZE];

read_simulate(int mb){
  struct file filep;
  struct inode ind;
  struct dentry dent;

  filep.f_pos = 0;
  filep.f_reada = 0;
  ind.i_rdev = (MAJOR_NR << 8);  /* Simulate an IDE disk */
  filep.f_dentry = &dent;
  dent.d_inode = &ind;

  do{
    block_read(&filep, test_buffer2, sizeof(test_buffer2), &filep.f_pos);
    jiffies++;
    if((filep.f_pos % 10000) == 0) printk("%d ", filep.f_pos);
  } while(filep.f_pos < mb*1000*1024);
}

/*
 * Simulate writes to a disk.  This will be cached, and ultimately the
 * cache will be flushed to the device.
 */
write_simulate(int mb){
  struct file filep;
  struct inode ind;
  struct dentry dent;

  filep.f_pos = 0;
  filep.f_reada = 0;
  ind.i_rdev = (MAJOR_NR << 8);  /* Simulate an IDE disk */
  filep.f_dentry = &dent;
  dent.d_inode = &ind;

  do{
    block_write(&filep, test_buffer2, sizeof(test_buffer2), &filep.f_pos);
    jiffies++;
    if((filep.f_pos % 10000) == 0) printk("%d ", filep.f_pos);
  } while(filep.f_pos < mb*1000*1024);

  sys_sync();
}

tape_write_simulate(int mb){
  struct file filep;
  struct inode ind;
  struct dentry dent;
  int rtn;

  filep.f_pos = 0;
  filep.f_reada = 0;
  ind.i_rdev = (CHR_MAJOR_NR << 8);  /* Simulate an IDE disk */
  filep.f_dentry = &dent;
  dent.d_inode = &ind;

  filep.f_mode = FMODE_WRITE;
  filep.f_flags = O_RDWR;
  filep.f_pos = 0;
  chrdev_open(&ind, &filep);

  do{
    rtn = (filep.f_op->write)(&filep, test_buffer2, sizeof(test_buffer2), &filep.f_pos);
    jiffies++;
    if((filep.f_pos % 10000) == 0) printk("%d ", filep.f_pos);
  } while(filep.f_pos < mb*1000*1024);

  return 0;
}

tape_read_simulate(int mb){
  struct file filep;
  struct inode ind;
  struct dentry dent;
  int rtn;

  filep.f_pos = 0;
  filep.f_reada = 0;
  ind.i_rdev = (CHR_MAJOR_NR << 8);  /* Simulate an IDE disk */
  filep.f_dentry = &dent;
  dent.d_inode = &ind;

  filep.f_mode = FMODE_READ;
  filep.f_flags = O_RDWR;
  filep.f_pos = 0;
  chrdev_open(&ind, &filep);

  do{
    rtn = (filep.f_op->read)(&filep, test_buffer2, sizeof(test_buffer2), &filep.f_pos);
    jiffies++;
    if((filep.f_pos % 10000) == 0) printk("%d ", filep.f_pos);
  } while(filep.f_pos < mb*1000*1024);
}

ioctl_simulate(){
  struct file filep;
  struct inode ind;
  struct dentry dent;
  int rtn;
  int i;

  filep.f_pos = 0;
  filep.f_reada = 0;
  ind.i_rdev = (CHR_MAJOR_NR << 8);  /* Simulate an IDE disk */
  filep.f_dentry = &dent;
  dent.d_inode = &ind;

  filep.f_mode = FMODE_WRITE;
  filep.f_flags = O_RDWR;
  filep.f_pos = 0;
  chrdev_open(&ind, &filep);

  for(i=0; i < 10; i++)
    {
      rtn = (filep.f_op->ioctl)(&ind, &filep, 
				SCSI_IOCTL_DOORLOCK, 0);
      rtn = (filep.f_op->ioctl)(&ind, &filep, 
				SCSI_IOCTL_DOORUNLOCK, 0);
  };
}
#if 0
swap_simulate()
{
    int i, j;
    int bzone[8];
    int zones;
    char buff[4096];

    zones = 8;
    for(i=0; i < 100; i++)
    {
	for(j=0; j < 8; j++)
	{
	    bzone[j] = i + j;
	}
	ll_rw_swap_file(WRITE, MAJOR_NR << 8, bzone, zones, buff);
    }

			      
}
#endif

/* 
 * Create a kernel thread.  We fake this with our own non-preemtive
 * scheduling, as this is the way it works in real life.  Debugging
 * with multiple kernel threads is really hard.
 */

_thread_create(void (*thr_func)(), 
	       struct task_struct * task,	       
	       unsigned int * jpnt)
{
    int i;
    static struct fs_struct allfs;
    unsigned int * tssptr;

    tssptr = (unsigned int *) &task->tss;
    for(i=0; i < 6; i++)
    {
	tssptr[i] = jpnt[i];
    }

    task->fs = &allfs;
    setup_tasklist((char *) task);
}

_thread_destroy(struct task_struct * this_task)
{
  int i;
  for(i=0; i < max_reg_tasks; i++)
    {
      if( task[i] == this_task )
	{
	  task[i] = NULL;
	  schedule();
	}
    }
}

void setup_tasklist(char * crnt)
{
  static struct task_struct * this_task;

  this_task = (struct task_struct *) crnt;

  this_task->addr_limit.seg = 0x50000000;

  task[max_reg_tasks++] = this_task;
}

int __down_trylock(struct semaphore * sem)
{
	return waking_non_zero_trylock(sem);
}

struct ctl_table_header * register_sysctl_table(ctl_table * table, 
			     int insert_at_head)
{
  table->de = (struct proc_dir_entry *) malloc(sizeof(struct proc_dir_entry));
}

_my_mem_init(void)
{
  int i;
  char * ptr = buffer0;

  ptr = (char *) (((unsigned int) ptr + 4095) & ~4095);
  buffer2 = ptr;
  for(i=0; i < NR_PAGES; i++)
    {
      freelist[i] = ptr;
      ptr += PAGE_SIZE;
    }
  next_free = NR_PAGES - 1;
}

int __map_nr(void * addr)
{
  return ((char *) addr - (char *) buffer2) >> 12;
}


int blk_ioctl(kdev_t dev, unsigned int cmd, unsigned long arg)
{
  return 0;
}

void kill_fasync(struct fasync_struct *fa, int sig)
{
}

int fasync_helper(int a, struct file * b, int c, struct fasync_struct **d)
{
}

