/*
 * These are the stub functions for ksim that are compiled against
 * normal user headers.  Anything that uses stdio, for example.
 */
#ifdef USE_VARARGS
#include <varargs.h>
#else
#include <stdarg.h>
#endif

#include <getopt.h>

#include <sys/mman.h>
#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <sys/time.h>

char opt = '\0';
extern int idle_thread;
extern int max_reg_tasks;

struct wait_queue
{
  int xxx;
};

void thread_wrapper(void (*thr_func)(void *), 
		    void * this_task,
		    void * arg)
{
  thr_func(arg);
  _thread_destroy((struct task_struct *) this_task);
}

thread_create(void *thr_func, void * arg)
{
    unsigned long * stack;
    struct task_struct * task;
    __jmp_buf jpnt;

    /*
     * First allocate a stack.  
     */
    stack = (char *) malloc(16384);
    task = (struct task_struct *) (((unsigned int) stack + 8192) & ~8191);
    memset(stack, 0, 16384);

    stack = (unsigned long *) ((unsigned int) task + 8188);
    *stack-- = arg;
    *stack-- = task;
    *stack-- = thr_func;

    jpnt[JB_PC] = (int) thread_wrapper;
    jpnt[JB_SP] = (int) stack;
    jpnt[JB_BP] = (int) stack;
    jpnt[JB_DI] = 0;
    jpnt[JB_SI] = 0;
    jpnt[JB_BX] = 0;
    _thread_create(thr_func, task, jpnt);
}

void schedule_oops(){
  fprintf(stderr,"Schedule");
  exit(1);
}

#ifdef USE_VARARGS
void panic(va_alist)
  va_dcl
#else
void panic(const char *format, ...)
#endif
{
  va_list vp;
#ifdef USE_VARARGS
  char *format;
#endif

  fprintf(stderr, "kbench: ");

#ifdef USE_VARARGS
  va_start(vp);
  format = va_arg (vp, char *);
#else
  va_start(vp, format);
#endif
  vfprintf(stderr, format, vp);
  va_end(va);

  fputc('\n', stderr);

  exit(1);
}

#ifdef USE_VARARGS
void printk(va_alist)
  va_dcl
#else
void printk(const char *format, ...)
#endif
{
  va_list vp;
#ifdef USE_VARARGS
  char *format;
#endif

#ifdef USE_VARARGS
  va_start(vp);
  format = va_arg (vp, char *);
#else
  va_start(vp, format);
#endif
  vfprintf(stdout, format, vp);
  va_end(va);
}

#if 0
void wake_up(struct wait_queue ** foo){
  if((*foo) == NULL) return;
  fprintf(stderr,"Wake up!\n");
}
#endif

#if 0
struct sigaction segv_act;

static void k_fault(int signal, struct sigcontext_struct context)
{
  struct sigcontext_struct *scp = &context;
  unsigned char * instr;

  instr = (unsigned char *) scp->eip;
    switch(*instr)
    {
    case 0xfa: /* cli() */
    case 0xfb: /* sti() */
      scp->eip++;
      break;
    case 0xec: /* inb al,dx */
      scp->eax = 0xff;
      scp->eip++;
      break;
    case 0x0f: /* movl %cr3,%eax  - used for invalidate */
      scp->eip += 3;
      break;
    case 0xed: /* in ax,dx */
      scp->eax = 0xff;
      scp->eip++;  
      break;
    case 0x88: /* movb %al, 0x0  - WP test in kernel */
      scp->eip += 6;
      break;
    default:
      printf("Unhandled segmentation fault\n");
      _exit(0);
    }
  return;
}
#endif

#if 0
struct sigcontext_struct {
	unsigned short sc_gs, __gsh;
	unsigned short sc_fs, __fsh;
	unsigned short sc_es, __esh;
	unsigned short sc_ds, __dsh;
	unsigned long sc_edi;
	unsigned long sc_esi;
	unsigned long sc_ebp;
	unsigned long sc_esp;
	unsigned long sc_ebx;
	unsigned long sc_edx;
	unsigned long sc_ecx;
	unsigned long sc_eax;
	unsigned long sc_trapno;
	unsigned long sc_err;
	unsigned long sc_eip;
	unsigned short sc_cs, __csh;
	unsigned long sc_efl;
	unsigned long esp_at_signal;
	unsigned short sc_ss, __ssh;
	unsigned long i387;
	unsigned long oldmask;
	unsigned long cr2;
};
#endif


void proc_unregister(void)
{
}

void * kmalloc(size_t size, int flags)
{
  return (void *) malloc(size);
}

void kfree(const void * objp)
{
  free(objp);
}

struct kmem_cache_s
{
  int alloc_size;
};

typedef struct kmem_cache_s kmem_cache_t;

void
kmem_cache_free(kmem_cache_t *cachep, void *objp)
{
  free(objp);
}

void *
kmem_cache_alloc(kmem_cache_t *cachep, int flags)
{
  return (void *) malloc(cachep->alloc_size);
}

kmem_cache_t *
kmem_cache_create(const char *name, size_t size, size_t offset,
	unsigned long flags, void (*ctor)(void*, kmem_cache_t *, unsigned long),
	void (*dtor)(void*, kmem_cache_t *, unsigned long))
{
  kmem_cache_t * rtn;
  rtn = (kmem_cache_t *) malloc(sizeof(*rtn));
  rtn->alloc_size = size;
  return rtn;
}

void grow_stack()
{
  char stackbuf[16384];
  memset(stackbuf, 0, sizeof(stackbuf));
}

int usage(void)
{
    fprintf(stderr, "help - display this information\n");
    fprintf(stderr, "exit - you figure it out\n");
    fprintf(stderr, "log loglevel - specify logging level\n");
    fprintf(stderr, "twrite - simulate tape reads\n");
    fprintf(stderr, "tread - simulate tape writes\n");
    fprintf(stderr, "read  - simulate disk reads\n");
    fprintf(stderr, "write - simulate disk writes\n");
    fprintf(stderr, "ioctl - simulate ioctls\n");
    fprintf(stderr, "swap  - simulate swap activity\n");
    fprintf(stderr, "\n");
}

main(int argc, char * argv[])
{
  unsigned int mem_start, mem_end, low_mem_start;
  extern char * buffer;
  extern unsigned int scsi_logging_level;
  FILE * zfile;
  char * current_task;
  int status;
  char linebuf[128];
  char * ptr;
  void * handles[100];
  int next_handle = 0;
  extern void  bdflush(void);
  extern void  idle(void);
  char c;
  int i;

  memset(handles, 0, sizeof(handles));

  /*
   * Make sure that we have memory allocated for the task struct.
   */
  grow_stack();

  _my_mem_init();

  /*
   * Buried somewhere in stackbuf will be the properly aligned structure
   * representing "current".
   */
  buffer = (char *) &i;
  current_task = (char *) ((unsigned int) buffer & ~8191);

  setup_tasklist(current_task);
  thread_create(bdflush, NULL);
  file_table_init();
  buffer_init();
  device_setup();
  idle_thread = max_reg_tasks;
  thread_create(idle, NULL);

  while(1==1)
  {
      fputs("ksim>", stdout);
      fgets(linebuf, sizeof(linebuf), stdin);

      /*
       * Get rid of trailing \n.
       */
      ptr = linebuf + strlen(linebuf) - 1;
      *ptr = '\0';

      if( strcmp(linebuf, "help") == 0 )
      {
	  usage();
	  continue;
      }

      if( strcmp(linebuf, "exit")  == 0 )
      {
	  break;
      }

      if( strcmp(linebuf, "quit")  == 0 )
      {
	  break;
      }

      if( strncmp(linebuf, "log", 3) == 0)
      {
	  ptr = linebuf + 4;
	  while(*ptr == ' ' || *ptr == '\t' ) ptr++;
	  scsi_logging_level = strtol(ptr, NULL, 16);
	  continue;
      }
      if( strcmp(linebuf, "write") == 0)
      {
	  opt = 'w'; /* Hack so that sleep_on schedules bdflush() */
	  fprintf(stderr, "Simulating disk write from dev_t == 0x800\n");
	  write_simulate(30);
	  continue;
      }
      if( strcmp(linebuf, "read") == 0 )
      {
	  struct timeval start, end;
	  unsigned int dsec, dusec;
	  double gsec;
	  fprintf(stderr, "Simulating disk read from dev_t == 0x800\n");
	  gettimeofday(&start, NULL);
	  read_simulate(30);
	  gettimeofday(&end, NULL);
	  dsec  = end.tv_sec  - start.tv_sec;
	  dusec = end.tv_usec - start.tv_usec;
	  dusec = dusec + dsec * 1000000;
	  gsec = dusec / 1000000.;
	  fprintf(stderr, "\nThroughput %g Mb/sec\n", 30.0/gsec);
	  continue;
      }
      if( strcmp(linebuf, "tread") == 0)
      {
	  fprintf(stderr, "Simulating tape read from dev_t == 0x900\n");
	  tape_read_simulate(10);
	  continue;
      }
      if( strcmp(linebuf, "twrite") == 0)
      {
	  fprintf(stderr, "Simulating tape write to dev_t == 0x900\n");
	  tape_write_simulate(10);
	  continue;
      }
      if( strcmp(linebuf, "ioctl") == 0)
      {
	  fprintf(stderr, "Simulating ioctl to dev_t == 0x800\n");
	  ioctl_simulate();
	  continue;
      }
#if 0
      if( strcmp(linebuf, "swap") == 0 )
      {
	  fprintf(stderr, "Swap to dev_t == 0x800\n");
	  swap_simulate();
	  continue;
      }
#endif
      if( strncmp(linebuf, "scsi_debug", 10) == 0 )
      {
	  scsi_debug_proc_info(linebuf, NULL, 0, strlen(linebuf), 
			       0, 1);
	  continue;
      }
      if( strncmp(linebuf, "scsi", 4) == 0 )
      {
	  scsi_proc_info(linebuf, NULL, 0, strlen(linebuf), 
			       0, 1);
	  continue;
      }
      if( strncmp(linebuf, "load", 4) == 0 )
      {
	   void * handle;
	   extern void * scsi_debug_get_handle(void);
	   handles[next_handle] = scsi_debug_get_handle();
	   scsi_register_module(1, handles[next_handle]);
	   next_handle++;
	   continue;
      }
      if( strncmp(linebuf, "unload", 6) == 0 )
      {
	  int mod_no;

	  mod_no = atol(linebuf + 7);
	  if( handles[mod_no] )
	  {
	      scsi_unregister_module(1, handles[mod_no]);
	      handles[mod_no] = NULL;
	  }
	  continue;
      }
      usage();
  }
}
void __pollwait()
{
  fprintf(stderr, "__pollwait called\n");
}

void fput(void)
{
  fprintf(stderr, "fput called\n");
}
